import os
import time

import pytest

from win32_setctime import setctime, SUPPORTED


def getctime(filepath):
    # Fix for Python 3.5 not supporting pathlib
    return os.path.getctime(str(filepath))


def test_supported():
    assert SUPPORTED


def test_setctime(tmp_path):
    filepath = tmp_path / "test_setctime.txt"
    timestamp = 946681200
    filepath.touch()
    setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


def test_directory(tmp_path):
    dirpath = tmp_path / "temp_dir"
    timestamp = 826618926
    dirpath.mkdir()
    setctime(dirpath, timestamp)
    assert getctime(dirpath) == timestamp


def test_timestamp_negative(tmp_path):
    filepath = tmp_path / "test_timestamp_negative.txt"
    timestamp = -5694948000
    filepath.touch()
    setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


def test_timestamp_with_nanoseconds(tmp_path):
    filepath = tmp_path / "test_timestamp_with_nanoseconds.txt"
    timestamp = 737206464.123456789
    filepath.touch()
    setctime(filepath, timestamp)
    assert getctime(filepath) == pytest.approx(timestamp)


def test_timestamp_lower_bound(tmp_path):
    filepath = tmp_path / "test_timestamp_lower_sdfds.dds"
    timestamp = -11644473599
    filepath.touch()
    setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


def test_timestamp_exceeds_lower_bound(tmp_path):
    filepath = tmp_path / "test_timestamp_exceeds_lower_bound.txt"
    timestamp = -11644473600
    filepath.touch()
    with pytest.raises(ValueError):
        setctime(filepath, timestamp)


def test_timestamp_far_in_the_future(tmp_path):
    filepath = tmp_path / "test_timestamp_far_in_the_future.txt"
    timestamp = 64675581821
    filepath.touch()
    setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


def test_timestamp_exceeds_upper_bound(tmp_path):
    filepath = tmp_path / "test_timestamp_exceeds_upper_bound.txt"
    timestamp = 1833029933770.9551616
    filepath.touch()
    with pytest.raises(ValueError):
        setctime(filepath, timestamp)


def test_file_does_not_exist(tmp_path):
    filepath = tmp_path / "test_file_does_not_exist.txt"
    timestamp = 123456789
    with pytest.raises(FileNotFoundError):
        setctime(filepath, timestamp)


def test_file_already_opened_read(tmp_path):
    filepath = tmp_path / "test_file_already_opened_read.txt"
    timestamp = 123456789
    filepath.touch()
    with open(str(filepath), "r"):
        setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


def test_file_already_opened_write(tmp_path):
    filepath = tmp_path / "test_file_already_opened_write.txt"
    timestamp = 123456789
    with open(str(filepath), "w"):
        setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


def test_file_already_opened_exclusive(tmp_path):
    filepath = tmp_path / "test_file_already_opened_write.txt"
    timestamp = 123456789
    with open(str(filepath), "x"):
        setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


def test_file_unicode(tmp_path):
    filepath = tmp_path / "𤭢.txt"
    timestamp = 123456789
    filepath.touch()
    setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


def test_forward_slash(tmp_path):
    folder = tmp_path / "foo" / "bar" / "baz"
    filepath = folder / "test_forward_slash.txt"
    timestamp = 123456789
    folder.mkdir(exist_ok=True, parents=True)
    filepath.touch()
    setctime(str(filepath).replace(r"\\", "/"), timestamp)
    assert getctime(filepath) == timestamp


def test_mtime_not_modified(tmp_path):
    filepath = tmp_path / "test_mtime_not_modified.txt"
    filepath.touch()
    before = os.path.getmtime(str(filepath))
    time.sleep(0.1)
    setctime(filepath, 123456789)
    assert os.path.getmtime(str(filepath)) == before


def test_atime_not_modified(tmp_path):
    filepath = tmp_path / "test_atime_not_modified.txt"
    filepath.touch()
    before = os.path.getatime(str(filepath))
    time.sleep(0.1)
    setctime(filepath, 123456789)
    assert os.path.getatime(str(filepath)) == before


def test_fix_file_tunneling(tmp_path):
    filepath = tmp_path / "test_fix_file_tunneling.txt"
    timestamp = 123456789
    filepath.touch()
    before = getctime(filepath)
    time.sleep(0.1)
    os.remove(str(filepath))
    filepath.touch()
    assert getctime(filepath) == before
    setctime(filepath, timestamp)
    assert getctime(filepath) == timestamp


@pytest.mark.parametrize("follow_symlinks", [True, False])
def test_with_symlinks(tmp_path, follow_symlinks):
    target_path = tmp_path / "target.txt"
    symlink_path = tmp_path / "symlink.txt"
    timestamp = 123456789
    target_path.touch()
    time.sleep(0.1)
    symlink_path.symlink_to(target_path)
    target_ctime = target_path.lstat().st_ctime
    symlink_ctime = symlink_path.lstat().st_ctime

    assert target_ctime != symlink_ctime

    setctime(symlink_path, timestamp, follow_symlinks=follow_symlinks)

    if follow_symlinks:
        assert symlink_path.lstat().st_ctime == symlink_ctime
        assert target_path.lstat().st_ctime == timestamp
    else:
        assert symlink_path.lstat().st_ctime == timestamp
        assert target_path.lstat().st_ctime == target_ctime



import os
import sys

try:
    from ctypes import byref, get_last_error, wintypes, FormatError, WinDLL, WinError

    kernel32 = WinDLL("kernel32", use_last_error=True)

    CreateFileW = kernel32.CreateFileW
    SetFileTime = kernel32.SetFileTime
    CloseHandle = kernel32.CloseHandle

    CreateFileW.argtypes = (
        wintypes.LPWSTR,
        wintypes.DWORD,
        wintypes.DWORD,
        wintypes.LPVOID,
        wintypes.DWORD,
        wintypes.DWORD,
        wintypes.HANDLE,
    )
    CreateFileW.restype = wintypes.HANDLE

    SetFileTime.argtypes = (
        wintypes.HANDLE,
        wintypes.PFILETIME,
        wintypes.PFILETIME,
        wintypes.PFILETIME,
    )
    SetFileTime.restype = wintypes.BOOL

    CloseHandle.argtypes = (wintypes.HANDLE,)
    CloseHandle.restype = wintypes.BOOL
except (ImportError, AttributeError, OSError, ValueError):
    SUPPORTED = False
else:
    SUPPORTED = os.name == "nt"


if sys.version_info >= (3, 6):
    PathLike = os.PathLike
else:
    from pathlib import PurePath as PathLike

from typing import Union


def setctime(
    filepath: Union[str, PathLike], timestamp: float, *, follow_symlinks: bool = True
) -> None:
    """Set the "ctime" (creation time) attribute of a file given an unix timestamp (Windows only)."""
    if not SUPPORTED:
        raise OSError("This function is only available for the Windows platform.")

    filepath = os.path.normpath(os.path.abspath(str(filepath)))
    timestamp = int(timestamp * 10000000) + 116444736000000000

    if not 0 < timestamp < (1 << 64):
        raise ValueError("The system value of the timestamp exceeds u64 size: %d" % timestamp)

    atime = wintypes.FILETIME(0xFFFFFFFF, 0xFFFFFFFF)
    mtime = wintypes.FILETIME(0xFFFFFFFF, 0xFFFFFFFF)
    ctime = wintypes.FILETIME(timestamp & 0xFFFFFFFF, timestamp >> 32)

    flags = 128 | 0x02000000

    if not follow_symlinks:
        flags |= 0x00200000

    handle = wintypes.HANDLE(CreateFileW(filepath, 256, 0, None, 3, flags, None))
    if handle.value == wintypes.HANDLE(-1).value:
        raise WinError(get_last_error())

    if not wintypes.BOOL(SetFileTime(handle, byref(ctime), byref(atime), byref(mtime))):
        raise WinError(get_last_error())

    if not wintypes.BOOL(CloseHandle(handle)):
        raise WinError(get_last_error())



import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open("src/win32_setctime/__init__.py", "r") as file:
    regex_version = r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]'
    version = re.search(regex_version, file.read(), re.MULTILINE).group(1)

with open("README.md", "rb") as file:
    readme = file.read().decode("utf-8")

setup(
    name="win32_setctime",
    version=version,
    packages=["win32_setctime"],
    package_dir={"": "src"},
    package_data={"win32_setctime": ["py.typed"]},
    description="A small Python utility to set file creation time on Windows",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Delgan",
    author_email="delgan.py@gmail.com",
    url="https://github.com/Delgan/win32-setctime",
    download_url="https://github.com/Delgan/win32-setctime/archive/{}.tar.gz".format(version),
    keywords=["win32", "windows", "filesystem", "filetime"],
    license="MIT license",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: System :: Filesystems",
        "Intended Audience :: Developers",
        "Environment :: Win32 (MS Windows)",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    extras_require={
        "dev": [
            "black>=19.3b0 ; python_version>='3.6'",
            "pytest>=4.6.2",
        ]
    },
    python_requires=">=3.5",
)
